from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from factrainer.base.dataset import IsImportableInstance
from numpy import typing as npt
from typing_extensions import TypeIs

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl


@runtime_checkable
class PdDataFrameProtocol(Protocol): ...


@runtime_checkable
class PdSeriesProtocol[T](Protocol): ...


@runtime_checkable
class PlDataFrameProtocol(Protocol): ...


@runtime_checkable
class PlSeriesProtocol(Protocol): ...


if TYPE_CHECKING:
    type PdDataFrameLike = pd.DataFrame
    type PdSeriesLike[T] = pd.Series[T]
    type PlDataFrameLike = pl.DataFrame
    type PlSeriesLike = pl.Series
else:
    type PdDataFrameLike = PdDataFrameProtocol
    type PdSeriesLike[T] = PdSeriesProtocol[T]
    type PlDataFrameLike = PlDataFrameProtocol
    type PlSeriesLike = PlSeriesProtocol

type MatrixLike = npt.NDArray[Any] | PdDataFrameLike | PlDataFrameLike
type VectorLike = npt.NDArray[Any] | PdSeriesLike[Any] | PlSeriesLike


class IsPdDataFrame(IsImportableInstance[PdDataFrameLike]):
    def is_instance(self, obj: Any) -> TypeIs[PdDataFrameLike]:
        try:
            import pandas as pd

            return isinstance(obj, pd.DataFrame)
        except ImportError:
            return False


class IsPdSeries(IsImportableInstance[PdSeriesLike[Any]]):
    def is_instance(self, obj: Any) -> TypeIs[PdSeriesLike[Any]]:
        try:
            import pandas as pd

            return isinstance(obj, pd.Series)
        except ImportError:
            return False


class IsPlDataFrame(IsImportableInstance[PlDataFrameLike]):
    def is_instance(self, obj: Any) -> TypeIs[PlDataFrameLike]:
        try:
            import polars as pl

            return isinstance(obj, pl.DataFrame)
        except ImportError:
            return False


class IsPlSeries(IsImportableInstance[PlSeriesLike]):
    def is_instance(self, obj: Any) -> TypeIs[PlSeriesLike]:
        try:
            import polars as pl

            return isinstance(obj, pl.Series)
        except ImportError:
            return False



from __future__ import annotations

from collections.abc import Generator
from typing import Any

import numpy as np
import polars as pl
from factrainer.base.dataset import IndexableDataset, RowIndex, Rows
from pydantic import field_validator

from sklearn.model_selection._split import _BaseKFold

from .types import (
    IsPdDataFrame,
    IsPdSeries,
    IsPlDataFrame,
    IsPlSeries,
    MatrixLike,
    VectorLike,
)


class SklearnDataset(IndexableDataset):
    X: MatrixLike
    y: VectorLike | None = None

    def k_fold_split(
        self, k_fold: _BaseKFold
    ) -> Generator[tuple[RowIndex, RowIndex], None, None]:
        for train_index, val_index in k_fold.split(self.X):
            yield train_index, val_index

    def __getitem__(self, index: Rows) -> "SklearnDataset":
        return SklearnDataset(
            X=self._getitem_X(self.X, index), y=self._getitem_y(self.y, index)
        )

    def _getitem_X(self, X: MatrixLike, index: Rows) -> MatrixLike:
        match index:
            case int():
                if isinstance(X, np.ndarray):
                    return np.expand_dims(X[index], axis=0)
                elif IsPdDataFrame().is_instance(X):
                    return X.take([index])
                elif IsPlDataFrame().is_instance(X):
                    raise NotImplementedError
                else:
                    raise ValueError
            case list():
                if isinstance(X, np.ndarray):
                    return X[index]
                elif IsPdDataFrame().is_instance(X):
                    return X.take(index)
                elif IsPlDataFrame().is_instance(X):
                    return (
                        pl.DataFrame({"__polars_row_index__": index})
                        .with_row_index(name="__factrainer_row_index__")
                        .join(
                            X.with_row_index(name="__polars_row_index__"),
                            on="__polars_row_index__",
                        )
                        .sort("__factrainer_row_index__")
                        .select(
                            pl.exclude(
                                ["__factrainer_row_index__", "__polars_row_index__"]
                            )
                        )
                    )
                else:
                    raise ValueError
            case np.ndarray():
                if isinstance(X, np.ndarray):
                    return X[index]
                elif IsPdDataFrame().is_instance(X):
                    return X.take(index)
                elif IsPlDataFrame().is_instance(X):
                    return (
                        pl.DataFrame({"__polars_row_index__": index})
                        .with_row_index(name="__factrainer_row_index__")
                        .join(
                            X.with_row_index(name="__polars_row_index__"),
                            on="__polars_row_index__",
                        )
                        .sort("__factrainer_row_index__")
                        .select(
                            pl.exclude(
                                ["__factrainer_row_index__", "__polars_row_index__"]
                            )
                        )
                    )
                else:
                    raise ValueError
            case slice():
                return X[index]
            case tuple():
                raise NotImplementedError
            case _:
                raise NotImplementedError

    def _getitem_y(self, y: VectorLike | None, index: Rows) -> VectorLike | None:
        if y is None:
            return None
        match index:
            case int():
                if isinstance(y, np.ndarray):
                    return np.expand_dims(y[index], axis=0) if y.ndim == 1 else y[index]
                elif IsPdSeries().is_instance(y):
                    return y.take([index])
                elif IsPlSeries().is_instance(y):
                    raise NotImplementedError
                else:
                    raise ValueError
            case list():
                if isinstance(y, np.ndarray):
                    return y[index]
                elif IsPdSeries().is_instance(y):
                    return y.take(index)
                elif IsPlSeries().is_instance(y):
                    return (
                        pl.DataFrame({"__polars_row_index__": index})
                        .with_row_index(name="__factrainer_row_index__")
                        .join(
                            pl.DataFrame(y).with_row_index(name="__polars_row_index__"),
                            on="__polars_row_index__",
                        )
                        .sort("__factrainer_row_index__")
                        .select(
                            pl.exclude(
                                ["__factrainer_row_index__", "__polars_row_index__"]
                            )
                        )
                        .to_series()
                    )
                else:
                    raise ValueError
            case np.ndarray():
                if isinstance(y, np.ndarray):
                    return y[index]
                elif IsPdSeries().is_instance(y):
                    return y.take(index)
                elif IsPlSeries().is_instance(y):
                    return (
                        pl.DataFrame({"__polars_row_index__": index})
                        .with_row_index(name="__factrainer_row_index__")
                        .join(
                            pl.DataFrame(y).with_row_index(name="__polars_row_index__"),
                            on="__polars_row_index__",
                        )
                        .sort("__factrainer_row_index__")
                        .select(
                            pl.exclude(
                                ["__factrainer_row_index__", "__polars_row_index__"]
                            )
                        )
                        .to_series()
                    )
                else:
                    raise ValueError
            case slice():
                return y[index]
            case _:
                raise TypeError

    @field_validator("X", mode="after")
    @classmethod
    def validate_X(cls, value: Any) -> MatrixLike:
        if isinstance(value, np.ndarray):
            return value
        if IsPdDataFrame().is_instance(value):
            return value
        if IsPlDataFrame().is_instance(value):
            return value
        raise ValueError(
            "X must be a numpy array, pandas DataFrame, or polars DataFrame"
        )

    @field_validator("y", mode="after")
    @classmethod
    def validate_y(cls, value: Any) -> VectorLike | None:
        if value is None:
            return None
        if isinstance(value, np.ndarray):
            return value
        if IsPdSeries().is_instance(value):
            return value
        if IsPlSeries().is_instance(value):
            return value
        raise ValueError("y must be a numpy array, pandas Series, or polars Series")



from copy import deepcopy
from enum import Enum, auto
from typing import Self

from factrainer.base.config import (
    BaseLearner,
    BasePredictConfig,
    BasePredictor,
    BaseTrainConfig,
    MlModelConfig,
)
from factrainer.base.dataset import Prediction
from pydantic import ConfigDict

from .dataset.dataset import SklearnDataset
from .raw_model import Predictable, ProbPredictable, SklearnModel


class SklearnTrainConfig(BaseTrainConfig):
    estimator: Predictable | ProbPredictable
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")


class SklearnPredictMethod(Enum):
    AUTO = auto()
    PREDICT = auto()
    PREDICT_PROBA = auto()


class SklearnPredictConfig(BasePredictConfig):
    predict_method: SklearnPredictMethod = SklearnPredictMethod.AUTO
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")


class SklearnLearner(BaseLearner[SklearnDataset, SklearnModel, SklearnTrainConfig]):
    def train(
        self,
        train_dataset: SklearnDataset,
        val_dataset: SklearnDataset | None,
        config: SklearnTrainConfig,
    ) -> SklearnModel:
        if train_dataset.y is None:
            raise ValueError("y cannot be None")
        _config = deepcopy(config)
        _config.estimator.fit(
            train_dataset.X,
            train_dataset.y,
            **(config.model_extra if config.model_extra is not None else {}),
        )
        return SklearnModel(estimator=_config.estimator)


class SklearnPredictor(
    BasePredictor[SklearnDataset, SklearnModel, SklearnPredictConfig]
):
    def predict(
        self,
        dataset: SklearnDataset,
        raw_model: SklearnModel,
        config: SklearnPredictConfig,
    ) -> Prediction:
        match config.predict_method:
            case SklearnPredictMethod.PREDICT_PROBA:
                if not hasattr(raw_model.estimator, "predict_proba"):
                    raise ValueError("The model does not support predict_proba method.")
                return raw_model.estimator.predict_proba(
                    dataset.X,
                    **(config.model_extra if config.model_extra is not None else {}),
                )
            case SklearnPredictMethod.PREDICT:
                if not hasattr(raw_model.estimator, "predict"):
                    raise ValueError("The model does not support predict method.")
                return raw_model.estimator.predict(
                    dataset.X,
                    **(config.model_extra if config.model_extra is not None else {}),
                )
            case SklearnPredictMethod.AUTO:
                if hasattr(raw_model.estimator, "predict_proba"):
                    return raw_model.estimator.predict_proba(
                        dataset.X,
                        **(
                            config.model_extra if config.model_extra is not None else {}
                        ),
                    )
                elif hasattr(raw_model.estimator, "predict"):
                    return raw_model.estimator.predict(
                        dataset.X,
                        **(
                            config.model_extra if config.model_extra is not None else {}
                        ),
                    )
                else:
                    raise ValueError(
                        "The model is not a valid classifier or regressor."
                    )
            case _:
                raise ValueError("Invalid predict method")


class SklearnModelConfig(
    MlModelConfig[
        SklearnDataset, SklearnModel, SklearnTrainConfig, SklearnPredictConfig
    ]
):
    learner: SklearnLearner
    predictor: SklearnPredictor
    train_config: SklearnTrainConfig
    pred_config: SklearnPredictConfig

    @classmethod
    def create(
        cls,
        train_config: SklearnTrainConfig,
        pred_config: SklearnPredictConfig | None = None,
    ) -> Self:
        return cls(
            learner=SklearnLearner(),
            predictor=SklearnPredictor(),
            train_config=train_config,
            pred_config=pred_config
            if pred_config is not None
            else SklearnPredictConfig(),
        )



from typing import Any, Protocol, Self, runtime_checkable

from factrainer.base.dataset import Prediction
from factrainer.base.raw_model import RawModel
from numpy import typing as npt

from .dataset.types import MatrixLike, VectorLike


@runtime_checkable
class BaseEstimatorProtocol(Protocol):
    def get_params(self, deep: bool = True) -> dict[str, Any]: ...
    def set_params(self, **params: Any) -> Self: ...


@runtime_checkable
class Predictable(BaseEstimatorProtocol, Protocol):
    def fit(
        self, X: MatrixLike, y: VectorLike | None, *args: Any, **kwargs: Any
    ) -> Self: ...

    def predict(self, X: npt.ArrayLike, *args: Any, **kwargs: Any) -> Prediction: ...


@runtime_checkable
class ProbPredictable(BaseEstimatorProtocol, Protocol):
    def fit(
        self, X: MatrixLike, y: VectorLike | None, *args: Any, **kwargs: Any
    ) -> Self: ...

    def predict_proba(
        self, X: npt.ArrayLike, *args: Any, **kwargs: Any
    ) -> Prediction: ...


class SklearnModel(RawModel):
    estimator: Predictable | ProbPredictable



from .config import (
    SklearnModelConfig,
    SklearnPredictConfig,
    SklearnPredictMethod,
    SklearnTrainConfig,
)
from .dataset.dataset import SklearnDataset
from .raw_model import SklearnModel

__all__ = [
    "SklearnPredictConfig",
    "SklearnTrainConfig",
    "SklearnModelConfig",
    "SklearnModel",
    "SklearnDataset",
    "SklearnPredictMethod",
]
